from typing import List, Optional, Dict, Any
from models.SequenceModel import SequenceModel


## @class DiagonalFarmDetectorModel
# Returns a list with (1) the total number of farms, (2) the length of the longest consecutive diagonal
# of farms, and (3) the number of farms with at least one diagonal neighbor, but zero adjacent neighbors
# Rationale: The ratio of the first two values gives a reasonable first impression of how much the player
# is using diagonals. If the third value is high (or really, if it's much above zero), it gives very strong
# evidence the player is using the diagonal strategy.
# @param levels: Levels applicable for model
class DiagonalFarmDetectorModel(SequenceModel):
    def __init__(self, levels: List[int] = []):
        '''
        @class DiagonalFarmDetectorModel
        Returns a list with (1) the total number of farms, (2) the length of the longest consecutive
        diagonal of farms, and (3) the number of farms with at least one diagonal neighbor, but zero
        adjacent neighbors
        :param levels: Levels applicable for model
        '''
        super().__init__()

    def _eval(self, events: List[Dict[str, Any]], verbose: bool = False) -> List[int]:
        assert events
        farmcount = 0
        diagonals = []
        diagonals_only = 0
        for event in reversed(events):
            # gamestate
            if event["event_custom"] == 0:
                tile_string = event["event_data_complex"]["tiles"]
                _tile = self._read_stringified_array(tile_string)
                tile = self._array_to_mat(4, _tile)
                for i in range(0, len(tile)):
                    if tile[i][3] == 9:
                        farmcount += 1
                        diagonals.append(self._right_diagonal(i, tile))
                        diagonals.append(self._left_diagonal(i, tile))
                        if self._diagonal_neighbors_present(i, tile) and not self._adjacent_neighbors_present(i, tile):
                            diagonals_only += 1
                break
        if diagonals:
            biggest_diagonal = max(diagonals)
        else:
            biggest_diagonal = 0
        return [farmcount, biggest_diagonal, diagonals_only]

    def _right_diagonal(self, index: int, tile: List[List[int]]) -> int:
        next = index + 51
        count = 1
        while next % 50 != 0:
            if tile[next][3] == 9:
                count += 1
                next += 51
            else:
                break
        return count

    def _left_diagonal(self, index: int, tile: List[List[int]]) -> int:
        next = index + 49
        count = 1
        while next % 50 != 49:
            if tile[next][3] == 9:
                count += 1
                next += 49
            else:
                break
        return count

    def _diagonal_neighbors_present(self, index: int, tile: List[List[int]]) -> bool:
        # check all four possibilities if it's not in the first or last row
        if 49 < index < 2450:
            if index % 50 != 0:
                if tile[index + 49][3] == 9 or tile[index - 51][3] == 9:
                    return True
            if index % 50 != 49:
                if tile[index + 51][3] == 9 or tile[index - 49][3] == 9:
                    return True
        # check both possibilities if it's in the first row
        elif index <= 49:
            if index != 0:
                if tile[index + 49][3] == 9:
                    return True
            if index != 49:
                if tile[index + 51][3] == 9:
                    return True
        # check both possibilities if it's in the last row
        elif index >= 2450:
            if index % 50 != 0:
                if tile[index - 51][3] == 9:
                    return True
            if index % 50 != 49:
                if tile[index - 49][3] == 9:
                    return True
        return False

    def _adjacent_neighbors_present(self, index: int, tile: List[List[int]]) -> bool:
        if index > 49:
            if tile[index - 50][3] == 9:
                return True
        if index < 2450:
            if tile[index + 50][3] == 9:
                return True
        if index % 50 != 0:
            if tile[index - 1][3] == 9:
                return True
        if index % 50 != 49:
            if tile[index + 1][3] == 9:
                return True
        return False




    # reformat raw variable functions
    def _read_stringified_array(self, arr: str) -> List[int]:
        if not arr:
            return []
        return [int(x) for x in arr.split(',')]


    def _array_to_mat(self, num_columns: int, arr: List[int]) -> List[List[int]]:
        assert len(arr) % num_columns == 0
        return [arr[i:i + num_columns] for i in range(0, len(arr), num_columns)]


def __repr__(self):
        return f"TownCompositionModel(levels={self._levels}, input_type={self._input_type})"


# import libraries
from ogd.core.models import Event
import typing
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class TimeToAnswerMS(Feature):
    def __init__(self, params:GeneratorParameters):
        Feature.__init__(self, params=params)
        self._latest_complete_lvl8  = None
        self._latest_complete_lvl16 = None
        self._latest_answer_Q0 = None
        self._latest_answer_Q2 = None
        self._answer_time = None

    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["CUSTOM.3", "COMPLETE.0"]
        # return ["QUESTION_ANSWER"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "COMPLETE.0":
            if event.GameState['level'] == 8:
                self._latest_complete_lvl8 = event.Timestamp
            if event.GameState['level'] == 16:
                self._latest_complete_lvl16 = event.Timestamp
        elif event.EventName == "CUSTOM.3":
            q_num = event.EventData["question"]
            if (q_num == 0):
                self._latest_answer_Q0 = event.Timestamp
            elif (q_num == 2):
                self._latest_answer_Q2 = event.Timestamp
            if self.CountIndex == q_num:
                self._answer_time = self._calcAnswerTime(timestamp=event.Timestamp)

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return ["Not Implemented"]

    # *** Optionally override public functions. ***

    # *** Other local functions
    def _calcAnswerTime(self, timestamp) -> Optional[int]:
        millis: Optional[float]
        if self.CountIndex == 0:
            millis = 1000.0 * (timestamp - self._latest_complete_lvl8).total_seconds()
        elif self.CountIndex == 1:
            millis = 1000.0 * (timestamp - self._latest_answer_Q0).total_seconds()
        elif self.CountIndex == 2:
            millis = 1000.0 * (timestamp - self._latest_complete_lvl16).total_seconds()
        elif self.CountIndex == 3:
            millis = 1000.0 * (timestamp - self._latest_answer_Q2).total_seconds()
        else:
            millis = None
        return int(millis) if millis is not None else None


from collections import defaultdict
from datetime import timedelta
from typing import Any, Dict, List

from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class PopulationSummary(SessionFeature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._user_sessions = defaultdict(list)
        self._user_completions = defaultdict(list)
        self._user_session_times = defaultdict(list)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["JobsCompleted", "SessionID", "SessionDuration"]

    def _updateFromEvent(self, event: Event) -> None:
        return

    def _updateFromFeatureData(self, feature:FeatureData):
        if feature.FeatureType == "JobsCompleted":
            self._user_completions[feature.PlayerID].append(feature.FeatureValues[0])
        elif feature.FeatureType == "SessionID" and feature.SessionID not in self._user_sessions[feature.PlayerID]:
            self._user_sessions[feature.PlayerID].append(feature.SessionID)
        elif feature.FeatureType == "SessionDuration":
            self._user_session_times[feature.PlayerID].append(feature.FeatureValues[0])

    def _getFeatureValues(self) -> List[Any]:
        num_sessions = [len(self._user_sessions[user]) for user in self._user_sessions]
        num_completions = [len(self._user_completions[user]) for user in self._user_completions]
        session_times : List[timedelta] = []

        for user in self._user_session_times:
            for time in self._user_session_times[user]:
                if type(time) == str and time == "No events":
                    pass
                else:
                    session_times.append(time)

        summary : Dict[str, float]= {
            "avg_session_count": 0,
            "avg_jobs_completed": 0,
            "avg_session_time": 0
        }

        if len(num_sessions) > 0:
            summary["avg_session_count"] = int(sum(num_sessions) / len(num_sessions))

        if len(num_completions) > 0:
            summary["avg_jobs_completed"] = int(sum(num_completions) / len(num_completions))

        if len(session_times) > 0:
            summary["avg_session_time"] = PopulationSummary._timesum(session_times) / len(session_times)

        return [summary]

    # *** Optionally override public functions. ***

    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        """List of ExtractionMode supported by the Feature.

        Overridden from base Feature version.
        A PlayerSummary is only used at player and population levels; not concerned with session-level.
        :return: _description_
        :rtype: List[ExtractionMode]
        """
        return [ExtractionMode.POPULATION]

    # *** Private Functions ***

    @staticmethod
    def _timesum(times:List[timedelta]):
        ret_val = timedelta(0)
        for time in times:
            ret_val += time
        return ret_val.total_seconds()



# import libraries
import json
from typing import Any, List, Optional
# import local files
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature




class SnippetReplace(PerLevelFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        PerLevelFeature.__init__(self, params=params)

        # out of time count
        self.sn_replace_num : int = 0
        #low attribute count
        self._notebook_click_num : int = 0
        # >>> create/initialize any variables to track feature extractor state <<<
        #
        # e.g. To track whether extractor found a click event yet:
        # self._found_click : bool = False

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return ["level_fail"] # >>> fill in names of events this Feature should use for extraction. <<<

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        """_summary_

        :return: _description_
        :rtype: List[str]
        """
        return [] # >>> fill in names of first-order features this Feature should use for extraction. <<<

    def _updateFromEvent(self, event:Event) -> None:
        """_summary_

        :param event: _description_
        :type event: Event
        """
        # >>> use the data in the Event object to update state variables as needed. <<<
        # Note that this function runs once on each Event whose name matches one of the strings returned by _eventFilter()
        #
        # e.g. check if the event name contains the substring "Click," and if so set self._found_click to True
       
        if("0" in event.EventData["fail_types"]):
            self._oot_count += 1
        else:
            self._low_attr_count += 1
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        """_summary_

        :param feature: _description_
        :type feature: FeatureData
        """
        # >>> use data in the FeatureData object to update state variables as needed. <<<
        # Note: This function runs on data from each Feature whose name matches one of the strings returned by _featureFilter().
        #       The number of instances of each Feature may vary, depending on the configuration and the unit of analysis at which this CustomFeature is run.
        return []

    def _getFeatureValues(self) -> List[Any]:
        """_summary_

        :return: _description_
        :rtype: List[Any]
        """
        
        # >>> use state variables to calculate the return value(s) of the base Feature and any Subfeatures. <<<
        # >>> put the calculated value(s) into a list as the function return value. <<<
        # >>> definitely don't return ["template"], unless you really find that useful... <<<
        #
        # e.g. use the self._found_click, which was created/initialized in __init__(...), and updated in _updateFromEvent(...):
        # if self._found_click:
        #     ret_val = [True]
        # else:
        #     ret_val = [False]
        # return ret_val
        #
        # note the code above is redundant, we could just return [self._found_click] to get the same result;
        # the more-verbose code is here for illustrative purposes.
        total_fails = self._oot_count + self._low_attr_count
        return [total_fails, self._oot_count, self._low_attr_count]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["OutOfTime", "LowAttribute"] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.POPULATION, ExtractionMode.PLAYER, ExtractionMode.SESSION] # >>> delete any modes you don't want run for your Feature. <<<
    
    # @staticmethod
    # def MinVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the minimum logging version for events to be processed by this Feature. <<<
    #     return super().MinVersion()

    # @staticmethod
    # def MaxVersion() -> Optional[str]:
    #     # >>> replace return statement below with a string defining the maximum logging version for events to be processed by this Feature. <<<
    #     return super().MaxVersion()



# import libraries
import json
from collections import defaultdict
from typing import Any, List
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData


class TopJobSwitchDestinations(Feature):

    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._current_session_id = None
        self._last_started_id = None
        self._mission_switch_pairs = defaultdict(dict)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["checkpoint"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        session_id = event.SessionID
        checkpoint = event.EventData["status"]
        mission_name = event.EventData["mission_id"]

        if checkpoint == "Begin Mission":
            if not self._last_started_id:
                self._last_started_id = mission_name
            elif session_id == self._current_session_id:
                # Started a new mission before completing the previous one
                if not mission_name in self._mission_switch_pairs[self._last_started_id].keys():
                    self._mission_switch_pairs[self._last_started_id][mission_name] = []

                self._mission_switch_pairs[self._last_started_id][mission_name].append(session_id) # here, we take what we switched to, and append where we switched from
        elif checkpoint == "Case Closed" and self._last_started_id:
            self._last_started_id = None

        # once we process the event, we know we're looking at data for this event's user next time.
        self._current_session_id = session_id

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        ret_val = {}

        for src in self._mission_switch_pairs.keys():
            dests = sorted(
                self._mission_switch_pairs[src].items(),
                key=lambda item: len(item[1]), # sort by length of list of ids.
                reverse=True # sort largest to smallest
            )
            ret_val[src] = {
                item[0]:item[1] for item in dests[0:5]
            }
            Logger.Log(f"For TopJobSwitchDestinations, sorted dests as: {json.dumps(dests)}")

        return [json.dumps(ret_val)]

    # *** Optionally override public functions. ***



# import libraries
import logging
from datetime import timedelta
from typing import Any, List
# import locals
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event, EventSource
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.utils.Logger import Logger

class SessionDuration(SessionFeature):

    def __init__(self, params:GeneratorParameters, threshold:int):
        self.threshold = threshold
        super().__init__(params=params)
        self.max_idle = timedelta(0)
        self.previous_time = None
        self.idle_time = timedelta(0)
        self.total_session_time = timedelta(0)
        # self._session_duration = 0
    def Subfeatures(self) -> List[str]:
            return ["Total", "Seconds", "Active", "ActiveSeconds", "Idle" "IdleSeconds", "MaxIdle"]
    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["all_events"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventSource == EventSource.GAME:
            if self.previous_time is not None:
                self.total_session_time += (event.Timestamp - self.previous_time)
                if (event.Timestamp - self.previous_time) > timedelta(minutes=1):
                    self.idle_time += (event.Timestamp - self.previous_time)
                    if self.max_idle < (event.Timestamp - self.previous_time):
                        self.max_idle = event.Timestamp - self.previous_time
            self.previous_time = event.Timestamp
     
        # # if this was earliest event, make it the start time.
        # if not self._client_start_time:
        #     self._client_start_time = event.Timestamp
        #     self._client_start_index = event.EventSequenceIndex
        # if self._client_start_time > event.Timestamp:
        #     Logger.Log(f"Got out-of-order events in SessionDuration for session {self._session_id};\nevent {event.EventName}:{event.EventSequenceIndex} for player {event.UserID}:{event.SessionID} had timestamp {event.Timestamp} earlier than start event, with time {self._client_start_time}, index {self._client_start_index}!", logging.WARN)
        #     self._client_start_time = event.Timestamp
        #     self._client_start_index = event.EventSequenceIndex
        # # if this was the latest event, make it the end time, otherwise output error.
        # if self._client_end_time is not None and self._client_end_time > event.Timestamp:
        #     Logger.Log(f"Got out-of-order events in SessionDuration for session {self._session_id};\nevent {event.EventName}:{event.EventSequenceIndex} for player {event.UserID}:{event.SessionID} had timestamp {event.Timestamp} earlier than end event, with time {self._client_end_time}, index {self._client_end_index}!", logging.WARN)
        # else:
        #     self._client_end_time = event.Timestamp
        #     self._client_end_index = event.EventSequenceIndex
        # # self._session_duration = (event.Timestamp - self._client_start_time).total_seconds()

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        if self.total_session_time is not None:
            return [self.total_session_time, self.total_session_time.total_seconds(), (self.total_session_time - self.idle_time), (self.total_session_time - self.idle_time).total_seconds(), self.idle_time, self.idle_time.total_seconds(), self.max_idle]
        else:
            return ["No events"]

    # *** Optionally override public functions. ***

    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.SESSION]



# import standard libraries
import logging
from typing import Any, Dict, List, Optional
# import local files
from ogd.core.schemas.games.AggregateSchema import AggregateSchema
from ogd.core.schemas.games.PerCountSchema import PerCountSchema
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class FeatureMapSchema(Schema):
    """
    Dumb struct to contain the specification and config of a set of features for a game.
    """
    def __init__(self, name:str, all_elements:Dict[str, Any]):
        self._legacy_mode           : bool
        self._legacy_perlevel_feats : Dict[str, PerCountSchema]
        self._percount_feats        : Dict[str, PerCountSchema]
        self._aggregate_feats       : Dict[str, AggregateSchema]

        if not isinstance(all_elements, dict):
            all_elements = {}
            Logger.Log(f"For FeatureMap config of `{name}`, all_elements was not a dict, defaulting to empty dict", logging.WARN)
        if "legacy" in all_elements.keys():
            self._legacy_mode = FeatureMapSchema._parseLegacyMode(legacy_element=all_elements['legacy'])
        else:
            Logger.Log(f"FeatureMap config does not have a 'legacy' element; defaulting to legacy=False", logging.WARN)
            self._legacy_mode = False
        if "perlevel" in all_elements.keys():
            self._legacy_perlevel_feats = FeatureMapSchema._parsePerLevelFeatures(perlevels=all_elements['perlevel'])
        else:
            Logger.Log(f"FeatureMap config does not have a 'perlevel' element; defaulting to empty dictionary", logging.WARN)
            self._legacy_perlevel_feats = {}
        if "per_count" in all_elements.keys():
            self._percount_feats = FeatureMapSchema._parsePerCountFeatures(percounts=all_elements['per_count'])
        else:
            Logger.Log(f"FeatureMap config does not have a 'per_count' element; defaulting to empty dictionary", logging.WARN)
            self._percount_feats = {}
        if "aggregate" in all_elements.keys():
            self._aggregate_feats = FeatureMapSchema._parseAggregateFeatures(aggregates=all_elements['aggregate'])
        else:
            Logger.Log(f"FeatureMap config does not have an 'aggregate' element; defaulting to empty dictionary", logging.WARN)
            self._aggregate_feats = {}
        _leftovers = { key : val for key,val in all_elements.items() if key not in {"legacy", "perlevel", "per_count", "aggregate"} }
        super().__init__(name=name, other_elements=_leftovers)

    @property
    def AsMarkdown(self) -> str:
        feature_summary = ["## Processed Features",
                           "The features/metrics calculated from this game's event logs by OpenGameData when an 'export' is run."
                          ]
        feature_list = [feature.AsMarkdown for feature in self._aggregate_feats.values()] + [feature.AsMarkdown for feature in self._percount_feats.values()]
        feature_list = feature_list if len(feature_list) > 0 else ["None"]
        return "  \n\n".join(feature_summary + feature_list)

    # @property
    # def AsMarkdownRow(self) -> str:
    #     ret_val : str = f"| {self.Name} | {self.ElementType} | {self.Description} |"
    #     if self.Details is not None:
    #         detail_markdowns = [f"**{name}** : {desc}" for name,desc in self.Details.items()]
    #         ret_val += ', '.join(detail_markdowns)
    #     ret_val += " |"
    #     return ret_val

    @property
    def LegacyMode(self) -> bool:
        return self._legacy_mode

    @property
    def LegacyPerLevelFeatures(self) -> Dict[str, PerCountSchema]:
        return self._legacy_perlevel_feats

    @property
    def PerCountFeatures(self) -> Dict[str, PerCountSchema]:
        return self._percount_feats

    @property
    def AggregateFeatures(self) -> Dict[str, AggregateSchema]:
        return self._aggregate_feats
    
    @staticmethod
    def _parseLegacyMode(legacy_element) -> bool:
        ret_val : bool
        if isinstance(legacy_element, dict):
            ret_val = legacy_element.get("enabled", False)
        else:
            ret_val = bool(legacy_element)
            Logger.Log(f"LegacyMode element was not a dict, defaulting to bool(legacy_element) == {ret_val}", logging.WARN)
        return ret_val
    
    @staticmethod
    def _parsePerLevelFeatures(perlevels) -> Dict[str, PerCountSchema]:
        ret_val : Dict[str, PerCountSchema]
        if isinstance(perlevels, dict):
            ret_val = { key : PerCountSchema(name=key, all_elements=val) for key,val in perlevels.items() }
        else:
            ret_val = {}
            Logger.Log(f"Per-level features map was not a dict, defaulting to empty dict", logging.WARN)
        return ret_val
    
    @staticmethod
    def _parsePerCountFeatures(percounts) -> Dict[str, PerCountSchema]:
        ret_val : Dict[str, PerCountSchema]
        if isinstance(percounts, dict):
            ret_val = { key : PerCountSchema(name=key, all_elements=val) for key,val in percounts.items() }
        else:
            ret_val = {}
            Logger.Log(f"Per-count features map was not a dict, defaulting to empty dict", logging.WARN)
        return ret_val
    
    @staticmethod
    def _parseAggregateFeatures(aggregates) -> Dict[str, AggregateSchema]:
        ret_val : Dict[str, AggregateSchema]
        if isinstance(aggregates, dict):
            ret_val = {key : AggregateSchema(name=key, all_elements=val) for key,val in aggregates.items()}
        else:
            ret_val = {}
            Logger.Log(f"Per-count features map was not a dict, defaulting to empty dict", logging.WARN)
        return ret_val



# import libraries
from mysql.connector import connection, cursor
import logging
import sshtunnel
import traceback
from datetime import datetime
from typing import Dict, Final, List, Tuple, Optional
# import locals
from ogd.core.interfaces.DataInterface import DataInterface
from ogd.core.models.enums.IDMode import IDMode
from ogd.core.schemas.configs.GameSourceSchema import GameSourceSchema
from ogd.core.schemas.configs.data_sources.MySQLSourceSchema import MySQLSchema
from ogd.core.utils.Logger import Logger


## @class SQL
#  A utility class containing some functions to assist in retrieving from a database.
#  Specifically, helps to connect to a database, make selections, and provides
#  a nicely formatted 500 error message.
class SQL:

    # Function to set up a connection to a database, via an ssh tunnel if available.
    @staticmethod
    def ConnectDB(schema:GameSourceSchema) -> Tuple[Optional[sshtunnel.SSHTunnelForwarder], Optional[connection.MySQLConnection]]:
        """
        Function to set up a connection to a database, via an ssh tunnel if available.

        :param db_settings: A dictionary mapping names of database parameters to values.
        :type db_settings: Dict[str,Any]
        :param ssh_settings: A dictionary mapping names of ssh parameters to values, or None if no ssh connection is desired., defaults to None
        :type ssh_settings: Optional[Dict[str,Any]], optional
        :return: A tuple consisting of the tunnel and database connection, respectively.
        :rtype: Tuple[Optional[sshtunnel.SSHTunnelForwarder], Optional[connection.MySQLConnection]]
        """
        ret_val : Tuple[Optional[sshtunnel.SSHTunnelForwarder], Optional[connection.MySQLConnection]] = (None, None)

        tunnel  : Optional[sshtunnel.SSHTunnelForwarder] = None
        db_conn : Optional[connection.MySQLConnection]   = None
        # Logger.Log("Preparing database connection...", logging.INFO)
        if schema.Source is not None and isinstance(schema.Source, MySQLSchema):
            if schema.Source.HasSSH:
                Logger.Log(f"Preparing to connect to MySQL via SSH, on host {schema.Source.SSH.Host}", level=logging.DEBUG)
                if (schema.Source.SSH.Host != "" and schema.Source.SSH.User != "" and schema.Source.SSH.Pass != ""):
                    tunnel,db_conn = SQL._connectToMySQLviaSSH(sql=schema.Source, db=schema.DatabaseName)
                else:
                    Logger.Log(f"SSH login had empty data, preparing to connect to MySQL directly instead, on host {schema.Source.DBHost}", level=logging.DEBUG)
                    db_conn = SQL._connectToMySQL(login=schema.Source, db=schema.DatabaseName)
                    tunnel = None
            else:
                Logger.Log(f"Preparing to connect to MySQL directly, on host {schema.Source.DBHost}", level=logging.DEBUG)
                db_conn = SQL._connectToMySQL(login=schema.Source, db=schema.DatabaseName)
                tunnel = None
            # Logger.Log("Done preparing database connection.", logging.INFO)
            ret_val = (tunnel, db_conn)
        else:
            Logger.Log(f"Unable to connect to MySQL, game source schema does not have a valid MySQL config!", level=logging.ERROR)

        return ret_val

    # Function to help connect to a mySQL server.
    @staticmethod
    def _connectToMySQL(login:MySQLSchema, db:str) -> Optional[connection.MySQLConnection]:
        """Function to help connect to a mySQL server.

        Simply tries to make a connection, and prints an error in case of failure.
        :param login: A SQLLogin object with the data needed to log into MySQL.
        :type login: SQLLogin
        :return: If successful, a MySQLConnection object, otherwise None.
        :rtype: Optional[connection.MySQLConnection]
        """
        try:
            Logger.Log(f"Connecting to SQL (no SSH) at {login.AsConnectionInfo}...", logging.DEBUG)
            db_conn = connection.MySQLConnection(host     = login.DBHost,    port    = login.DBPort,
                                                 user     = login.DBUser,    password= login.DBPass,
                                                 database = db, charset = 'utf8')
            Logger.Log(f"Connected.", logging.DEBUG)
            return db_conn
        #except MySQLdb.connections.Error as err:
        except Exception as err:
            msg = f"""Could not connect to the MySql database.
            Login info: {login.AsConnectionInfo} w/port type={type(login.DBPort)}.
            Full error: {type(err)} {str(err)}"""
            Logger.Log(msg, logging.ERROR)
            traceback.print_tb(err.__traceback__)
            return None

    ## Function to help connect to a mySQL server over SSH.
    @staticmethod
    def _connectToMySQLviaSSH(sql:MySQLSchema, db:str) -> Tuple[Optional[sshtunnel.SSHTunnelForwarder], Optional[connection.MySQLConnection]]:
        """Function to help connect to a mySQL server over SSH.

        Simply tries to make a connection, and prints an error in case of failure.
        :param sql: A SQLLogin object with the data needed to log into MySQL.
        :type sql: SQLLogin
        :param ssh: An SSHLogin object with the data needed to log into MySQL.
        :type ssh: SSHLogin
        :return: An open connection to the database if successful, otherwise None.
        :rtype: Tuple[Optional[sshtunnel.SSHTunnelForwarder], Optional[connection.MySQLConnection]]
        """
        tunnel    : Optional[sshtunnel.SSHTunnelForwarder] = None
        db_conn   : Optional[connection.MySQLConnection] = None
        MAX_TRIES : Final[int] = 5
        tries : int = 0
        connected_ssh : bool = False

        # First, connect to SSH
        while connected_ssh == False and tries < MAX_TRIES:
            if tries > 0:
                Logger.Log("Re-attempting to connect to SSH.", logging.INFO)
            try:
                Logger.Log(f"Connecting to SSH at {sql.SSHConfig.AsConnectionInfo}...", logging.DEBUG)
                tunnel = sshtunnel.SSHTunnelForwarder(
                    (sql.SSH.Host, sql.SSH.Port), ssh_username=sql.SSH.User, ssh_password=sql.SSH.Pass,
                    remote_bind_address=(sql.DBHost, sql.DBPort), logger=Logger.std_logger
                )
                tunnel.start()
                connected_ssh = True
                Logger.Log(f"Connected.", logging.DEBUG)
            except Exception as err:
                msg = f"Could not connect via SSH: {type(err)} {str(err)}"
                Logger.Log(msg, logging.ERROR)
                Logger.Print(msg, logging.ERROR)
                traceback.print_tb(err.__traceback__)
                tries = tries + 1
        if connected_ssh == True and tunnel is not None:
            # Then, connect to MySQL
            try:
                Logger.Log(f"Connecting to SQL (via SSH) at {sql.DBUser}@{sql.DBHost}:{tunnel.local_bind_port}/{db}...", logging.DEBUG)
                db_conn = connection.MySQLConnection(host     = sql.DBHost,    port    = tunnel.local_bind_port,
                                                     user     = sql.DBUser,    password= sql.DBPass,
                                                     database = db, charset ='utf8')
                Logger.Log(f"Connected", logging.DEBUG)
                return (tunnel, db_conn)
            except Exception as err:
                msg = f"Could not connect to the MySql database: {type(err)} {str(err)}"
                Logger.Log(msg, logging.ERROR)
                Logger.Print(msg, logging.ERROR)
                traceback.print_tb(err.__traceback__)
                if tunnel is not None:
                    tunnel.stop()
                return (None, None)
        else:
            return (None, None)

    @staticmethod
    def disconnectMySQL(db:Optional[connection.MySQLConnection], tunnel:Optional[sshtunnel.SSHTunnelForwarder]=None) -> None:
        if db is not None:
            db.close()
            Logger.Log("Closed MySQL database connection", logging.DEBUG)
        else:
            Logger.Log("No MySQL database to close.", logging.DEBUG)
        if tunnel is not None:
            tunnel.stop()
            Logger.Log("Stopped MySQL tunnel connection", logging.DEBUG)
        else:
            Logger.Log("No MySQL tunnel to stop", logging.DEBUG)

    # Function to build and execute SELECT statements on a database connection.
    @staticmethod
    def SELECT(cursor        :cursor.MySQLCursor,          db_name        : str,                   table    : str,
               columns       :List[str]           = [],    filter         : Optional[str] = None,
               sort_columns  :Optional[List[str]] = None,  sort_direction : str           = "ASC", grouping : Optional[str] = None,
               distinct      :bool                = False, offset         : int           = 0,     limit    : int           = -1,
               fetch_results :bool                = True,  params         : Tuple[str]    = tuple()) -> Optional[List[Tuple]]:
        """Function to build and execute SELECT statements on a database connection.

        :param cursor: A database cursor, retrieved from the active connection.
        :type cursor: cursor.MySQLCursor
        :param db_name: The name of the database to which we are connected.
        :type db_name: str
        :param table: The name of the table from which we want to make a selection.
        :type table: str
        :param columns: A list of columns to be selected. If empty (or None), all columns will be used (SELECT * FROM ...). Defaults to None
        :type columns: List[str], optional
        :param filter: A string giving the constraints for a WHERE clause (The "WHERE" term itself should not be part of the filter string), defaults to None
        :type filter: str, optional
        :param sort_columns: A list of columns to sort results on. The order of columns in the list is the order given to SQL. Defaults to None
        :type sort_columns: List[str], optional
        :param sort_direction: The "direction" of sorting, either ascending or descending., defaults to "ASC"
        :type sort_direction: str, optional
        :param grouping: A column name to group results on. Subject to SQL rules for grouping, defaults to None
        :type grouping: str, optional
        :param distinct: A bool to determine whether to select only rows with distinct values in the column, defaults to False
        :type distinct: bool, optional
        :param limit: The maximum number of rows to be selected. Use -1 for no limit., defaults to -1
        :type limit: int, optional
        :param fetch_results: A bool to determine whether all results should be fetched and returned, defaults to True
        :type fetch_results: bool, optional
        :return: A collection of all rows from the selection, if fetch_results is true, otherwise None.
        :rtype: Optional[List[Tuple]]
        """
        d          = "DISTINCT" if distinct else ""
        cols       = ",".join([f"{col}" for col in columns]) if len(columns) > 0 else "*"
        sort_cols  = ",".join([f"`{col}`" for col in sort_columns]) if sort_columns is not None and len(sort_columns) > 0 else None
        table_path = db_name + "." + str(table)

        sel_clause = f"SELECT {d} {cols} FROM {table_path}"
        where_clause = "" if filter    is None else f"WHERE {filter}"
        group_clause = "" if grouping  is None else f"GROUP BY {grouping}"
        sort_clause  = "" if sort_cols is None else f"ORDER BY {sort_cols} {sort_direction}"
        lim_clause   = "" if limit < 0         else f"LIMIT {str(max(offset, 0))}, {str(limit)}" # don't use a negative for offset
        query = f"{sel_clause} {where_clause} {group_clause} {sort_clause} {lim_clause};"
        return SQL.Query(cursor=cursor, query=query, params=params, fetch_results=fetch_results)

    @staticmethod
    def Query(cursor:cursor.MySQLCursor, query:str, params:Optional[Tuple], fetch_results: bool = True) -> Optional[List[Tuple]]:
        result : Optional[List[Tuple]] = None
        # first, we do the query.
        Logger.Log(f"Running query: {query}\nWith params: {params}", logging.DEBUG, depth=3)
        start = datetime.now()
        cursor.execute(query, params)
        time_delta = datetime.now()-start
        Logger.Log(f"Query execution completed, time to execute: {time_delta}", logging.DEBUG)
        # second, we get the results.
        if fetch_results:
            result = cursor.fetchall()
            time_delta = datetime.now()-start
            Logger.Log(f"Query fetch completed, total query time:    {time_delta} to get {len(result) if result is not None else 0:d} rows", logging.DEBUG)
        return result

class MySQLInterface(DataInterface):

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, game_id:str, config:GameSourceSchema, fail_fast:bool):
        self._tunnel    : Optional[sshtunnel.SSHTunnelForwarder] = None
        self._db        : Optional[connection.MySQLConnection] = None
        self._db_cursor : Optional[cursor.MySQLCursor] = None
        super().__init__(game_id=game_id, config=config, fail_fast=fail_fast)
        self.Open()

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    def _open(self, force_reopen:bool = False) -> bool:
        if force_reopen:
            self.Close()
            self.Open(force_reopen=False)
        if not self._is_open:
            start = datetime.now()
            if isinstance(self._config.Source, MySQLSchema):
                self._tunnel, self._db = SQL.ConnectDB(schema=self._config)
                if self._db is not None:
                    self._db_cursor = self._getCursor()
                    self._is_open = True
                    time_delta = datetime.now() - start
                    Logger.Log(f"Database Connection Time: {time_delta}", logging.INFO)
                    return True
                else:
                    Logger.Log(f"Unable to open MySQL interface.", logging.ERROR)
                    SQL.disconnectMySQL(tunnel=self._tunnel, db=self._db)
                    return False
            else:
                Logger.Log(f"Unable to open MySQL interface, the schema has invalid type {type(self._config)}", logging.ERROR)
                SQL.disconnectMySQL(tunnel=self._tunnel, db=self._db)
                return False
        else:
            return True

    def _close(self) -> bool:
        SQL.disconnectMySQL(tunnel=self._tunnel, db=self._db)
        Logger.Log("Closed connection to MySQL.", logging.DEBUG)
        self._is_open = False
        return True

    def _allIDs(self) -> List[str]:
        if self._db_cursor is not None and isinstance(self._config.Source, MySQLSchema):
            _db_name     : str = self._config.DatabaseName
            _table_name  : str = self._config.TableName

            sess_id_col  : str = self._TableSchema.SessionIDColumn or "session_id"

            filters : List[str] = []
            params  : List[str] = []
            if _table_name != self._game_id:
                filters.append(f"`app_id`=%s")
                params.append(self._game_id)
            filter_clause = " AND ".join(filters)
            
            data = SQL.SELECT(cursor =self._db_cursor, db_name=_db_name,      table   =_table_name,
                              columns=[sess_id_col],   filter =filter_clause, distinct=True,
                              params =tuple(params))
            return [str(id[0]) for id in data] if data != None else []
        else:
            Logger.Log(f"Could not get list of all session ids, MySQL connection is not open.", logging.WARN)
            return []

    def _fullDateRange(self) -> Dict[str,datetime]:
        ret_val = {'min':datetime.now(), 'max':datetime.now()}
        if self._db_cursor is not None and isinstance(self._config.Source, MySQLSchema):
            _db_name     : str = self._config.DatabaseName
            _table_name  : str = self._config.TableName

            # prep filter strings
            filters = []
            params  = []
            if _table_name != self._game_id:
                filters.append(f"`app_id`=%s")
                params.append(self._game_id)
            filter_clause = " AND ".join(filters)

            # run query
            result = SQL.SELECT(cursor=self._db_cursor, db_name=_db_name, table=_table_name,
                                columns=['MIN(server_time)', 'MAX(server_time)'], filter=filter_clause,
                                params =tuple(params))
            if result is not None:
                ret_val = {'min':result[0][0], 'max':result[0][1]}
        else:
            Logger.Log(f"Could not get full date range, MySQL connection is not open or config was not for MySQL.", logging.WARN)
        return ret_val

    def _rowsFromIDs(self, id_list:List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]]=None, exclude_rows:Optional[List[str]]=None) -> List[Tuple]:
        ret_val = []
        # grab data for the given session range. Sort by event time, so
        if self._db_cursor is not None and isinstance(self._config.Source, MySQLSchema):
            # filt = f"app_id='{self._game_id}' AND (session_id  BETWEEN '{next_slice[0]}' AND '{next_slice[-1]}'){ver_filter}"
            _db_name     : str = self._config.DatabaseName
            _table_name  : str = self._config.TableName

            sess_id_col = self._TableSchema.SessionIDColumn or 'session_id'
            play_id_col = self._TableSchema.UserIDColumn or 'player_id'
            seq_idx_col = self._TableSchema.EventSequenceIndexColumn or 'session_n'
            evt_nam_col = self._TableSchema.EventNameColumn or "event_name"

            filters = []
            params = []
            if _table_name != self._game_id:
                filters.append(f"`app_id`=%s")
                params.append(self._game_id)
            # if versions is not None and versions is not []:
            #     filters.append(f"app_version in ({','.join([str(version) for version in versions])})")
            id_param_string = ",".join( [f"%s"]*len(id_list) )
            if id_mode == IDMode.SESSION:
                filters.append(f"`{sess_id_col}` IN ({id_param_string})")
                params += [str(id) for id in id_list]
            elif id_mode == IDMode.USER:
                filters.append(f"`{play_id_col}` IN ({id_param_string})")
                params += [str(id) for id in id_list]
            else:
                raise ValueError("Invalid IDMode in MySQLInterface!")
            if exclude_rows is not None:
                evt_name_param_string = ",".join( ["%s"]*len(exclude_rows) )
                filters.append(f"`{evt_nam_col}` not in ({evt_name_param_string})")
                params += [str(name) for name in exclude_rows]
            filter_clause = " AND ".join(filters)

            data = SQL.SELECT(cursor=self._db_cursor, db_name=_db_name,                        table=_table_name,
                              filter=filter_clause,   sort_columns=[sess_id_col, seq_idx_col], sort_direction="ASC",
                              params=tuple(params))
            if data is not None:
                ret_val = data
            # self._select_queries.append(select_query) # this doesn't appear to be used???
        else:
            Logger.Log(f"Could not get data for {len(id_list)} sessions, MySQL connection is not open or config was not for MySQL.", logging.WARN)
        return ret_val

    def _IDsFromDates(self, min:datetime, max:datetime, versions:Optional[List[int]]=None) -> List[str]:
        ret_val = []
        if self._db_cursor is not None and isinstance(self._config.Source, MySQLSchema):
            # alias long setting names.
            _db_name     : str = self._config.DatabaseName
            _table_name  : str = self._config.TableName

            # prep filter strings
            filters = []
            params = []
            if _table_name != self._game_id:
                filters.append(f"`app_id`=%s")
                params.append(self._game_id)
            # if versions is not None and versions is not []:
            #     filters.append(f"app_version in ({','.join([str(version) for version in versions])})")
            filters.append(f"`{self._TableSchema.EventSequenceIndexColumn}`='0'")
            filters.append(f"(`server_time` BETWEEN '{min.isoformat()}' AND '{max.isoformat()}')")
            filter_clause = " AND ".join(filters)

            # run query
            # We grab the ids for all sessions that have 0th move in the proper date range.
            sess_id_col = self._TableSchema.SessionIDColumn or "`session_id`"
            sess_ids_raw = SQL.SELECT(cursor=self._db_cursor,   db_name=_db_name,     table=_table_name,
                                     columns=[sess_id_col],     filter=filter_clause,
                                     sort_columns=[sess_id_col], sort_direction="ASC", distinct=True,
                                     params=tuple(params))
            if sess_ids_raw is not None:
                ret_val = [str(sess[0]) for sess in sess_ids_raw]
        else:
            Logger.Log(f"Could not get session list for {min.isoformat()}-{max.isoformat()} range, MySQL connection is not open or config was not for MySQL.", logging.WARN)
        return ret_val

    def _datesFromIDs(self, id_list:List[str], id_mode:IDMode=IDMode.SESSION, versions:Optional[List[int]]=None) -> Dict[str, datetime]:
        ret_val = {'min':datetime.now(), 'max':datetime.now()}
        if self._db_cursor is not None and isinstance(self._config.Source, MySQLSchema):
            # alias long setting names.
            _db_name     : str = self._config.DatabaseName
            _table_name  : str = self._config.TableName
            
            # prep filter strings
            filters = []
            params = tuple()
            if _table_name != self._game_id:
                filters.append(f"`app_id`=%s")
                params = tuple(self._game_id)
            # if versions is not None and versions is not []:
            #     filters.append(f"app_version in ({','.join([str(version) for version in versions])})")
            ids_string = ','.join([f"'{x}'" for x in id_list])
            if id_mode == IDMode.SESSION:
                sess_id_col = self._TableSchema.SessionIDColumn or "session_id"
                filters.append(f"{sess_id_col} IN ({ids_string})")
            elif id_mode == IDMode.USER:
                play_id_col = self._TableSchema.UserIDColumn or "player_id"
                filters.append(f"`{play_id_col}` IN ({ids_string})")
            else:
                raise ValueError("Invalid IDMode in MySQLInterface!")
            filter_clause = " AND ".join(filters)
            # run query
            result = SQL.SELECT(cursor=self._db_cursor,      db_name=_db_name,    table=_table_name,
                                columns=['MIN(server_time)', 'MAX(server_time)'], filter=filter_clause,
                                params=params)
            if result is not None:
                ret_val = {'min':result[0][0], 'max':result[0][1]}
        else:
            Logger.Log(f"Could not get date range for {len(id_list)} sessions, MySQL connection is not open.", logging.WARN)
        return ret_val

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***

    def _getCursor(self) -> Optional[cursor.MySQLCursor]:
        ret_val : Optional[cursor.MySQLCursor] = None

        if self._db is not None:
            _cursor = self._db.cursor()
            if isinstance(_cursor, cursor.MySQLCursor):
                ret_val = _cursor
            else:
                Logger.Log(f"db.cursor() call returned a cursor of unexpected type {type(ret_val)}, can not access the database!", logging.ERROR)
        return ret_val



from typing import Any, List
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class PlayMode(SessionFeature):

    def __init__(self, params: GeneratorParameters):
        self.play_mode = None
        super().__init__(params=params)

    @classmethod
    def _getEventDependencies(cls, mode: ExtractionMode) -> List[str]:
        return ["game_start"]

    @classmethod
    def _getFeatureDependencies(cls, mode: ExtractionMode) -> List[str]:
        return []

    def _extractFromEvent(self, event: Event) -> None:
        if event.EventName == "game_start":
            mode = event.EventData.get("mode")
            if mode:
                self.play_mode = mode

    def _extractFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self.play_mode]



# import standard libraries

from argparse import ArgumentParser, SUPPRESS
from pathlib import Path
from typing import List

# import 3rd-party libraries

# import local files

class OGDParsers:
    @staticmethod
    def CommandParser(games_list:List[str]) -> ArgumentParser:
        game_parser = OGDParsers.GameParser(games_list=games_list) 
        export_parser = OGDParsers.ExportParser(parents=[game_parser])
        destination_parser = OGDParsers.DestinationParser()

        parser = ArgumentParser(description="Simple command-line utility to execute OpenGameData export requests.")
        sub_parsers = parser.add_subparsers(help="Chosen command to run", dest="command")
        sub_parsers.add_parser("export", parents=[export_parser, destination_parser],
                                help="Export data in a given date range.")
        sub_parsers.add_parser("export-events", parents=[export_parser, destination_parser],
                                help="Export event data in a given date range.")
        sub_parsers.add_parser("export-features", parents=[export_parser, destination_parser],
                                help="Export session feature data in a given date range.")
        sub_parsers.add_parser("info", parents=[game_parser],
                                help="Display info about the given game.")
        sub_parsers.add_parser("readme", parents=[game_parser, destination_parser],
                                help="Generate a readme for the given game.")
        sub_parsers.add_parser("list-games",
                                help="Display a list of games available for parsing.")
        # sub_parsers.add_parser("help",
        #                         help="Display a list of games available for parsing.")
        return parser

    @staticmethod
    def ExportParser(parents:List[ArgumentParser]) -> ArgumentParser:
        export_parser = ArgumentParser(add_help=False, parents=parents)
        export_parser.add_argument("start_date", nargs="?", default=None,
                            help="The starting date of an export range in MM/DD/YYYY format (defaults to today).")
        export_parser.add_argument("end_date", nargs="?", default=None,
                            help="The ending date of an export range in MM/DD/YYYY format (defaults to today).")
        export_parser.add_argument("-p", "--player", default="",
                            help="Tell the program to output data for a player with given ID, instead of using a date range.")
        export_parser.add_argument("-s", "--session", default="",
                            help="Tell the program to output data for a session with given ID, instead of using a date range.")
        export_parser.add_argument("--player_id_file", default="",
                            help="Tell the program to output data for a collection of players with IDs in given file, instead of using a date range.")
        export_parser.add_argument("--session_id_file", default="",
                            help="Tell the program to output data for a collection of sessions with IDs in given file, instead of using a date range.")
        export_parser.add_argument("-f", "--file", default="",
                            help="Tell the program to use a file as input, instead of looking up a database.")
        export_parser.add_argument("-m", "--monthly", default=False, action="store_true",
                            help="Set the program to export a month's-worth of data, instead of using a date range. Replace the start_date argument with a month in MM/YYYY format.")
        # allow individual feature files to be skipped.
        export_parser.add_argument("--no_session_file", default=False, action="store_true",
                            help="Tell the program to skip outputting a per-session file.")
        export_parser.add_argument("--no_player_file", default=False, action="store_true",
                            help="Tell the program to skip outputting a per-player file.")
        export_parser.add_argument("--no_pop_file", default=False, action="store_true",
                            help="Tell the program to skip outputting a population file.")
        return export_parser

    @staticmethod
    def DestinationParser() -> ArgumentParser:
        destination_parser = ArgumentParser(add_help=False)
        destination_parser.add_argument("-d", "--destination", default=SUPPRESS, type=str,
                            help="The destination folder for the readme.")
        return destination_parser

    @staticmethod
    def GameParser(games_list:List[str]) -> ArgumentParser:
        game_parser = ArgumentParser(add_help=False)
        game_parser.add_argument("game", type=str.upper, choices=games_list,
                            help="The game to use with the given command.")
        return game_parser


from typing import Any, List

from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class JobsCompleted(SessionFeature):

    def __init__(self, params:GeneratorParameters, player_id:str):
        self._player_id = player_id
        super().__init__(params=params)
        self._jobs_completed = []

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["complete_job"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event: Event) -> None:
        if event.UserID == self._player_id:
            _job_name = event.GameState.get('job_name', event.EventData.get('job_name', "JOB NAME NOT FOUND"))
            if event.app_version == 'Aqualab' or event.app_version == 'None':
                self._jobs_completed.append(_job_name.get('string_value'))
            else:
                self._jobs_completed.append(_job_name)

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [len(self._jobs_completed), self._jobs_completed]

    # *** Optionally override public functions. ***

    def Subfeatures(self) -> List[str]:
        return ["Names"]



import logging
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Optional, List
# import locals

class Logger:
    debug_level : int = logging.INFO
    std_logger  : logging.Logger   = logging.getLogger("std_logger")
    file_logger : Optional[logging.Logger] = None

    @staticmethod
    def InitializeLogger(level:int, use_logfile:bool):
        """Function to set up the stdout and file loggers of the Logging package.

        :param level: The logging level, which must be one of the levels defined by the Python `logging` package (`ERROR`, `WARN`, `INFO`, or `DEBUG`)
        :type level: int
        :param use_logfile: Bool for whether to use file output for logging or not.
        :type use_logfile: bool
        """
        # Set up loggers. First, the std out logger
        if not Logger.std_logger.hasHandlers():
            stdout_handler = logging.StreamHandler()
            Logger.std_logger.addHandler(stdout_handler)
            print("Added handler to std_logger")
        else:
            Logger.std_logger.warning(f"Trying to add a handler to std_logger, when handlers ({Logger.std_logger.handlers}) already exist!")
        _valid_levels = {logging.ERROR, logging.WARN, logging.WARNING, logging.INFO, logging.DEBUG}
        if level in _valid_levels:
            Logger.std_logger.setLevel(level=level)
        else:
            Logger.std_logger.setLevel(level=logging.INFO)
            Logger.std_logger.info("No valid logging level given, defaulting to INFO.")
        Logger.std_logger.info("Initialized standard out logger")

        # Then, set up the file logger. Check for permissions errors.
        if use_logfile:
            Logger.file_logger = logging.getLogger("file_logger")
            Logger.file_logger.setLevel(level=logging.DEBUG)
            # file_logger.setLevel(level=logging.DEBUG)
            try:
                err_handler = logging.FileHandler("./ExportErrorReport.log", encoding="utf-8")
                debug_handler = logging.FileHandler("./ExportDebugReport.log", encoding="utf-8")
            except PermissionError as err:
                Logger.std_logger.exception(f"Failed permissions check for log files. No file logging on server.")
            else:
                Logger.std_logger.info("Successfully set up logging files.")
                err_handler.setLevel(level=logging.WARNING)
                Logger.file_logger.addHandler(err_handler)
                debug_handler.setLevel(level=logging.DEBUG)
                Logger.file_logger.addHandler(debug_handler)
            finally:
                Logger.file_logger.debug("Initialized file logger")
    
    # Function to print a method to both the standard out and file logs.
    # Useful for "general" errors where you just want to print out the exception from a "backstop" try-catch block.
    @staticmethod
    def Log(message:str, level=logging.INFO, depth:int=0) -> None:
        now = datetime.now().strftime("%y-%m-%d %H:%M:%S")
        indent = ''.join(['  '*depth])
        _idt_msg = message.replace("\n", f"\n{' '*9}{indent}")
        if Logger.file_logger is not None:
            match level:
                case logging.DEBUG:
                    Logger.file_logger.debug(f"DEBUG:   {now} {indent}{_idt_msg}")
                case logging.INFO:
                    Logger.file_logger.info( f"INFO:    {now} {indent}{_idt_msg}")
                case logging.WARNING:
                    Logger.file_logger.warning( f"WARNING: {now} {indent}{_idt_msg}")
                case logging.ERROR:
                    Logger.file_logger.error(f"ERROR:   {now} {indent}{_idt_msg}")
        if Logger.std_logger is not None:
            match level:
                case logging.DEBUG:
                    Logger.std_logger.debug(f"DEBUG:   {indent}{_idt_msg}")
                case logging.INFO:
                    Logger.std_logger.info( f"INFO:    {indent}{_idt_msg}")
                case logging.WARNING:
                    Logger.std_logger.warning( f"WARNING: {indent}{_idt_msg}")
                case logging.ERROR:
                    Logger.std_logger.error(f"ERROR:   {indent}{_idt_msg}")

    @staticmethod
    def Print(message:str, level=logging.DEBUG) -> None:
        match level:
            case logging.DEBUG:
                print(f"debug:   {message}")
            case logging.INFO:
                print(f"info:    {message}")
            case logging.WARNING:
                print(f"warning: {message}")
            case logging.ERROR:
                print(f"error:   {message}")



from typing import List
from statistics import median
## import local libraries
from models.FeatureModel import FeatureModel


class PlayingTimeModel(FeatureModel):
    def __init__(self, features: List[str]):
        super().__init__()
        self._features = features

    def _eval(self, sess: dict, verbose: bool = False):
        res = {}
        if sess['continue'] or sess['debug']:
            return None
        for x in self._features:
            res[x] = sess[x]
        return res



# import standard libraries
import csv
import logging
import traceback
from argparse import Namespace
from itertools import chain
from pathlib import Path
from typing import Any, List, Optional, Set, Tuple

# import 3rd-party libraries

# import local files
from ogd.core.exec.Generators import OGDGenerators
from ogd.core.interfaces.DataInterface import DataInterface
from ogd.core.interfaces.CSVInterface import CSVInterface
from ogd.core.interfaces.outerfaces.DataOuterface import DataOuterface
from ogd.core.interfaces.outerfaces.TSVOuterface import TSVOuterface
from ogd.core.interfaces.outerfaces.DebugOuterface import DebugOuterface
from ogd.core.managers.ExportManager import ExportManager
from ogd.core.requests.RequestResult import RequestResult, ResultStatus
from ogd.core.requests.Request import Request, ExporterRange
from ogd.core.schemas.configs.ConfigSchema import ConfigSchema
from ogd.core.schemas.configs.GameSourceSchema import GameSourceSchema
from ogd.core.models.enums.ExportMode import ExportMode
from ogd.core.schemas.games.GameSchema import GameSchema
from ogd.core.models.enums.IDMode import IDMode
from ogd.core.schemas.tables.TableSchema import TableSchema
from ogd.core.utils.Logger import Logger
from ogd.core.utils.Readme import Readme

class OGDCommands:
    """Utility class to collect functions for each of the commands that can be run when executing ogd-core as a module.
    """
    @staticmethod
    def ListGames(games_list:List[str]) -> bool:
        print(f"The games available for export are:\n{games_list}")
        return True

    @staticmethod
    def ShowGameInfo(config:ConfigSchema, game:str) -> bool:
        """Function to print out info on a game from the game's schema.
    This does a similar function to writeReadme, but is limited to the CSV metadata part
    (basically what was in the schema, at one time written into the csv's themselves).
    Further, the output is printed rather than written to file.

        :return: True if game metadata was successfully loaded and printed, or False if an error occurred
        :rtype: bool
        """
        try:
            game_schema = GameSchema.FromFile(game_id=game)
            table_schema = TableSchema(schema_name=f"{config.GameSourceMap[game].TableSchema}.json")
            readme = Readme(game_schema=game_schema, table_schema=table_schema)
            print(readme.CustomReadmeSource)
        except Exception as err:
            msg = f"Could not print information for {game}: {type(err)} {str(err)}"
            Logger.Log(msg, logging.ERROR)
            traceback.print_tb(err.__traceback__)
            return False
        else:
            return True

    @staticmethod
    def WriteReadme(config:ConfigSchema, game:str, destination:Path) -> bool:
        """Function to write out the readme file for a given game.
    This includes the CSV metadata (data from the schema, originally written into
    the CSV files themselves), custom readme source, and the global changelog.
    The readme is placed in the game's data folder.

        :return: _description_
        :rtype: bool
        """
        path = destination / game
        try:
            game_schema = GameSchema.FromFile(game_id=game, schema_path=Path("src") / "ogd" / "games" / game / "schemas")
            table_schema = TableSchema(schema_name=f"{config.GameSourceMap[game].TableSchema}.json")
            readme = Readme(game_schema=game_schema, table_schema=table_schema)
            readme.GenerateReadme(path=path)
        except Exception as err:
            msg = f"Could not create a readme for {game}: {type(err)} {str(err)}"
            Logger.Log(msg, logging.ERROR)
            traceback.print_tb(err.__traceback__)
            return False
        else:
            Logger.Log(f"Successfully generated a readme for {game}.", logging.INFO)
            return True

    @staticmethod
    def RunExport(args:Namespace, config:ConfigSchema, destination:Path, with_events:bool = False, with_features:bool = False) -> bool:
        """Function to handle execution of export code.
        This is the main intended use of the program.

        :param events: _description_, defaults to False
        :type events: bool, optional
        :param features: _description_, defaults to False
        :type features: bool, optional
        :return: _description_
        :rtype: bool
        """
        success : bool = False

        export_modes   : Set[ExportMode]
        interface      : DataInterface
        range          : ExporterRange
        file_outerface : DataOuterface
        dataset_id     : Optional[str] = None

    # 1. get exporter modes to run
        export_modes = OGDGenerators.genModes(with_events=with_events, with_features=with_features,
                                              no_session_file=args.no_session_file, no_player_file=args.no_player_file, no_pop_file=args.no_pop_file)
    # 2. figure out the interface and range; optionally set a different dataset_id
        if args.file is not None and args.file != "":
            # raise NotImplementedError("Sorry, exports with file inputs are currently broken.")
            _ext = str(args.file).split('.')[-1]
            _cfg = GameSourceSchema(name="FILE SOURCE", all_elements={"schema":"OGD_EVENT_FILE"}, data_sources={})
            interface = CSVInterface(game_id=args.game, config=_cfg, fail_fast=config.FailFast, filepath=Path(args.file), delim="\t" if _ext == 'tsv' else ',')
            # retrieve/calculate id range.
            ids = interface.AllIDs()
            range = ExporterRange.FromIDs(source=interface, ids=ids if ids is not None else [])
        else:
            interface = OGDGenerators.genDBInterface(config=config, game=args.game)
        # a. Case where specific player ID was given
            if args.player is not None and args.player != "":
                range = ExporterRange.FromIDs(source=interface, ids=[args.player], id_mode=IDMode.USER)
                dataset_id = f"{args.game}_{args.player}"
        # b. Case where player ID file was given
            elif args.player_id_file is not None and args.player_id_file != "":
                file_path = Path(args.player_id_file)
                with open(file_path) as player_file:
                    reader = csv.reader(player_file)
                    file_contents = list(reader) # this gives list of lines, each line a list
                    names = list(chain.from_iterable(file_contents)) # so, convert to single list
                    print(f"list of names: {list(names)}")
                    range = ExporterRange.FromIDs(source=interface, ids=names, id_mode=IDMode.USER)
        # c. Case where specific session ID was given
            elif args.session is not None and args.session != "":
                range = ExporterRange.FromIDs(source=interface, ids=[args.session], id_mode=IDMode.SESSION)
                dataset_id = f"{args.game}_{args.session}"
        # d. Case where session ID file was given
            elif args.session_id_file is not None and args.session_id_file != "":
                file_path = Path(args.session_id_file)
                with open(file_path) as session_file:
                    reader = csv.reader(session_file)
                    file_contents = list(reader) # this gives list of lines, each line a list
                    names = list(chain.from_iterable(file_contents)) # so, convert to single list
                    print(f"list of sessions: {list(names)}")
                    range = ExporterRange.FromIDs(source=interface, ids=names, id_mode=IDMode.SESSION)
        # e. Default case where we use date range
            else:
                range = OGDGenerators.genDateRange(game=args.game, interface=interface, monthly=args.monthly, start_date=args.start_date, end_date=args.end_date)
    # 3. set up the outerface, based on the range and dataset_id.
        _cfg = GameSourceSchema(name="FILE DEST", all_elements={"database":"FILE", "table":"DEBUG", "schema":"OGD_EVENT_FILE"}, data_sources={})
        file_outerface = TSVOuterface(game_id=args.game, config=_cfg, export_modes=export_modes, date_range=range.DateRange,
                                    file_indexing=config.FileIndexConfig, dataset_id=dataset_id)
        outerfaces : Set[DataOuterface] = {file_outerface}
        # If we're in debug level of output, include a debug outerface, so we know what is *supposed* to go through the outerfaces.
        if config.DebugLevel == "DEBUG":
            _cfg = GameSourceSchema(name="DEBUG", all_elements={"database":"DEBUG", "table":"DEBUG", "schema":"OGD_EVENT_FILE"}, data_sources={})
            outerfaces.add(DebugOuterface(game_id=args.game, config=_cfg, export_modes=export_modes))

    # 4. Once we have the parameters parsed out, construct the request.
        req = Request(range=range, exporter_modes=export_modes, interface=interface, outerfaces=outerfaces)
        if req.Interface.IsOpen():
            export_manager : ExportManager = ExportManager(config=config)
            result         : RequestResult = export_manager.ExecuteRequest(request=req)
            success = result.Status == ResultStatus.SUCCESS
            level = logging.INFO if success else logging.ERROR
            Logger.Log(message=result.Message, level=level)
            Logger.Log(f"Total data request execution time: {result.Duration}", logging.INFO)
            # cProfile.runctx("feature_exporter.ExportFromSQL(request=req)",
                            # {'req':req, 'feature_exporter':feature_exporter}, {})
        return success


## import standard libraries
from typing import Any, Callable, Dict, List, Optional
## import local files
from . import features
from ogd.core.generators.detectors.Detector import Detector
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.GeneratorLoader import GeneratorLoader
from ogd.core.generators.extractors.Feature import Feature
from ogd.games.SHIPWRECKS.features import *
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.games.GameSchema import GameSchema

## @class ShipwrecksLoader
#  Extractor subclass for extracting features from Shipwrecks game data.
class ShipwrecksLoader(GeneratorLoader):
    ## Constructor for the ShipwrecksLoader class.
    #  Initializes some custom private data (not present in base class) for use
    #  when calculating some features.
    #  Sets the sessionID feature.
    #  Further, initializes all Q&A features to -1, representing unanswered questions.
    #
    #  @param session_id The id number for the session whose data is being processed
    #                    by this extractor instance.
    #  @param game_table A data structure containing information on how the db
    #                    table assiciated with this game is structured. 
    #  @param game_schema A dictionary that defines how the game data itself is
    #                     structured.
    def __init__(self, player_id:str, session_id:str, game_schema: GameSchema, mode:ExtractionMode, feature_overrides:Optional[List[str]]):
        super().__init__(player_id=player_id, session_id=session_id, game_schema=game_schema, mode=mode, feature_overrides=feature_overrides)

    @staticmethod
    def _getFeaturesModule():
        return features

    def _loadFeature(self, feature_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any]) -> Feature:
        ret_val : Feature
        if extractor_params._count_index is None:
            match feature_type:
                case "ActiveJobs":
                    ret_val = ActiveJobs.ActiveJobs(params=extractor_params)
                case "EventList":
                    ret_val = EventList.EventList(params=extractor_params)
                case "EvidenceBoardCompleteCount":
                    ret_val = EvidenceBoardCompleteCount.EvidenceBoardCompleteCount(params=extractor_params)
                case "JobsCompleted":
                    ret_val = JobsCompleted.JobsCompleted(params=extractor_params, session_id=self._session_id)
                case "PlayerSummary":
                    ret_val = PlayerSummary.PlayerSummary(params=extractor_params)
                case "PopulationSummary":
                    ret_val = PopulationSummary.PopulationSummary(params=extractor_params)
                case "SessionDuration":
                    ret_val = SessionDuration.SessionDuration(params=extractor_params, session_id=self._session_id)
                case "SessionID":
                    ret_val = SessionID.SessionID(params=extractor_params, session_id=self._session_id)
                case "TopJobCompletionDestinations":
                    ret_val = TopJobCompletionDestinations.TopJobCompletionDestinations(params=extractor_params)
                case "TopJobSwitchDestinations":
                    ret_val = TopJobSwitchDestinations.TopJobSwitchDestinations(params=extractor_params)
                case "TotalDiveTime":
                    ret_val = TotalDiveTime.TotalDiveTime(params=extractor_params)
                case _:
                    raise NotImplementedError(f"'{feature_type}' is not a valid aggregate feature for Shipwrecks.")
        else:
            match feature_type:
                case "MissionDiveTime":
                    ret_val = MissionDiveTime.MissionDiveTime(params=extractor_params)
                case "JobsAttempted":
                    ret_val = JobsAttempted.JobsAttempted(params=extractor_params, mission_map=self._game_schema.NonStandardElements["mission_map"])
                case "MissionSonarTimeToComplete":
                    ret_val = MissionSonarTimeToComplete.MissionSonarTimeToComplete(params=extractor_params)
                case _:
                    raise NotImplementedError(f"'{feature_type}' is not a valid per-count feature for Shipwrecks.")
        return ret_val

    def _loadDetector(self, detector_type:str, extractor_params:GeneratorParameters, schema_args:Dict[str,Any], trigger_callback:Callable[[Event], None]) -> Detector:
        raise NotImplementedError(f"'{detector_type}' is not a valid feature for Shipwrecks.")



# import libraries
import abc
from datetime import datetime, timedelta
from typing import Any, Dict, Optional
# import locals
from ogd.core.models.Event import Event, EventSource
from ogd.core.utils.typing import Map

class DetectorEvent(Event):
   def __init__(self, session_id:str,      app_id:str,
                event_name:str,            event_data:Map,
                timestamp:datetime,        time_offset:Optional[timedelta],
                app_version:Optional[str], log_version:Optional[str],
                user_id:Optional[str],     user_data:Optional[Map],
                game_state:Optional[Map],  event_sequence_index:Optional[int]):
      """Constructor for the simple DetectorEvent wrapper.

      :param session_id: _description_
      :type session_id: str
      :param app_id: _description_
      :type app_id: str
      :param event_name: _description_
      :type event_name: str
      :param event_data: _description_
      :type event_data: Map
      :param timestamp: _description_, defaults to datetime.now()
      :type timestamp: datetime, optional
      :param app_version: _description_, defaults to None
      :type app_version: Optional[str], optional
      :param log_version: _description_, defaults to None
      :type log_version: Optional[str], optional
      :param time_offset: _description_, defaults to None
      :type time_offset: Optional[int], optional
      :param user_id: _description_, defaults to ""
      :type user_id: Optional[str], optional
      :param user_data: _description_, defaults to {}
      :type user_data: Optional[Map], optional
      :param game_state: _description_, defaults to {}
      :type game_state: Optional[Map], optional
      :param event_sequence_index: _description_, defaults to None
      :type event_sequence_index: Optional[int], optional
      """
      super().__init__(session_id=session_id,   app_id=app_id,           timestamp=timestamp,
                       event_name=event_name,   event_data=event_data,   event_source=EventSource.GENERATED,
                       app_version=app_version, log_version=log_version, time_offset=time_offset,
                       user_id=user_id,         user_data=user_data,
                       game_state=game_state,   event_sequence_index=event_sequence_index)


# import libraries
import logging
from datetime import timedelta
from typing import Any, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.games.AQUALAB.features.PerJobFeature import PerJobFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class JobDiveTime(PerJobFeature):

    def __init__(self, params:GeneratorParameters, job_map:dict):
        super().__init__(params=params, job_map=job_map)
        self._session_id = None
        self._dive_start_time = None
        self._prev_timestamp = None
        self._time = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["all_events"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.SessionID != self._session_id:
            self._session_id = event.SessionID
            # if we jumped to a new session, we only want to count time up to last event, not the time between sessions.
            if self._dive_start_time and self._prev_timestamp:
                self._time += (self._prev_timestamp - self._dive_start_time).total_seconds()
                self._dive_start_time = event.Timestamp

        if event.EventName == "begin_dive":
            self._dive_start_time = event.Timestamp
        elif event.EventName == "scene_changed":
            if self._dive_start_time is not None:
                self._time += (event.Timestamp - self._dive_start_time).total_seconds()
                self._dive_start_time = None

        self._prev_timestamp = event.Timestamp

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [timedelta(seconds=self._time)]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



# standard imports
import logging
from datetime import date, datetime
from pathlib import Path
from typing import Any, Dict, Optional

# ogd imports
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

# local imports
from ogd.core.schemas.datasets.DatasetSchema import DatasetSchema

class FileListConfigSchema(Schema):
    def __init__(self, name:str, all_elements:Dict[str, Any]):
        self._files_base     : Optional[str]
        self._templates_base : Optional[str]
        if not isinstance(all_elements, dict):
            all_elements = {}
        if "files_base" in all_elements.keys():
            self._files_base = FileListConfigSchema._parseFilesBase(all_elements["files_base"])
        else:
            self._files_base = None
        if "templates_base" in all_elements.keys():
            self._templates_base = FileListConfigSchema._parseTemplatesBase(all_elements["templates_base"])
        else:
            self._templates_base = None
        _used = {"files_base", "templates_base"}
        _leftovers = { key : val for key,val in all_elements.items() if key not in _used }
        super().__init__(name=name, other_elements=_leftovers)

    def __str__(self) -> str:
        return str(self.Name)

    @property
    def FilesBase(self) -> Optional[str]:
        return self._files_base
    @property
    def TemplatesBase(self) -> Optional[str]:
        return self._templates_base
    @property
    def AsMarkdown(self) -> str:
        ret_val : str = self.Name
        return ret_val

    @staticmethod
    def EmptySchema() -> "FileListConfigSchema":
        return FileListConfigSchema(name="CONFIG NOT FOUND", all_elements={})

    @staticmethod
    def _parseFilesBase(files_base) -> str:
        ret_val : str
        if isinstance(files_base, str):
            ret_val = files_base
        else:
            ret_val = str(files_base)
            Logger.Log(f"Filepath base was unexpected type {type(files_base)}, defaulting to str(files_name)={ret_val}.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseTemplatesBase(templates_base) -> str:
        ret_val : str
        if isinstance(templates_base, str):
            ret_val = templates_base
        else:
            ret_val = str(templates_base)
            Logger.Log(f"Templates base was unexpected type {type(templates_base)}, defaulting to str(templates_name)={ret_val}.", logging.WARN)
        return ret_val

class GameDatasetCollectionSchema(Schema):
    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, name:str, all_elements:Dict[str, Any]):
        self._game_datasets : Dict[str, DatasetSchema]

        if not isinstance(all_elements, dict):
            all_elements = {}
        self._game_datasets = GameDatasetCollectionSchema._parseGameDatasets(datasets=all_elements)
        super().__init__(name=name, other_elements={})

    def __str__(self) -> str:
        return str(self.Name)

    # *** Properties ***

    @property
    def Datasets(self) -> Dict[str, DatasetSchema]:
        return self._game_datasets
    @property
    def AsMarkdown(self) -> str:
        ret_val : str = self.Name
        return ret_val

    @staticmethod
    def EmptySchema() -> "GameDatasetCollectionSchema":
        return GameDatasetCollectionSchema(name="DATASET COLLECTION NOT FOUND", all_elements={})

    @staticmethod
    def _parseGameDatasets(datasets:Dict[str, Any]) -> Dict[str, DatasetSchema]:
        ret_val : Dict[str, DatasetSchema]
        if isinstance(datasets, dict):
            ret_val = {
                key : DatasetSchema(key, dataset if isinstance(dataset, dict) else {})
                for key,dataset in datasets.items()
            }
        else:
            ret_val = {}
            Logger.Log(f"Collection of datasets was unexpected type {type(datasets)}, defaulting to empty dictionary.", logging.WARN)
        return ret_val

class FileListSchema(Schema):
    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, name:str, all_elements:Dict[str, Any]):
        self._games_file_lists : Dict[str, GameDatasetCollectionSchema]
        self._config           : FileListConfigSchema

        if not isinstance(all_elements, dict):
            all_elements = {}
    # 1. Parse config
        if "CONFIG" in all_elements.keys():
            self._config = FileListSchema._parseConfig(config=all_elements["CONFIG"])
        else:
            self._date_modified = "UNKNOWN"
    # 2. Parse games
        _used = {"CONFIG"}
        _leftovers = { key : val for key,val in all_elements.items() if key not in _used }
        self._games_file_lists = FileListSchema._parseGamesFileLists(games_dict=_leftovers)
        super().__init__(name=name, other_elements={})

    def __str__(self) -> str:
        return self.Name

    # *** Properties ***

    @property
    def Games(self) -> Dict[str, GameDatasetCollectionSchema]:
        return self._games_file_lists
    @property
    def Config(self) -> FileListConfigSchema:
        return self._config
    @property
    def AsMarkdown(self) -> str:
        ret_val : str = self.Name
        return ret_val

    # *** Private Functions ***

    # NOTE: Yes, most of these parse functions are redundant, but that's fine,
    # we just want to have one bit of code to parse each piece of the schema, even if most do the same thing.

    @staticmethod
    def _parseConfig(config) -> FileListConfigSchema:
        ret_val : FileListConfigSchema
        if isinstance(config, dict):
            ret_val = FileListConfigSchema(name="ConfigSchema", all_elements=config)
        else:
            ret_val = FileListConfigSchema.EmptySchema()
            Logger.Log(f"Config was unexpected type {type(config)}, defaulting to empty ConfigSchema.", logging.WARN)
        return ret_val

    @staticmethod
    def _parseGamesFileLists(games_dict:Dict[str, Any]) -> Dict[str, GameDatasetCollectionSchema]:
        ret_val : Dict[str, GameDatasetCollectionSchema]
        if isinstance(games_dict, dict):
            ret_val = {
                key : GameDatasetCollectionSchema(key, datasets if isinstance(datasets, dict) else {})
                for key, datasets in games_dict.items()
            }
        else:
            ret_val = {}
            Logger.Log(f"Collection of games was unexpected type {type(games_dict)}, defaulting to empty dictionary.", logging.WARN)
        return ret_val




# import standard libraries
from enum import IntEnum

class IterationMode(IntEnum):
    """Simple enum to represent the different kinds of extractor iteration:
    Aggregate (no iteration) or Per-Count (iterated by count)
    """
    AGGREGATE = 1
    PERCOUNT = 2

    def __str__(self):
        return self.name



from typing import List
from statistics import median
from datetime import datetime as dt
## import local libraries
from models.FeatureModel import FeatureModel

## @class TutorialProgressionModel
# Returns dictionary of time of first pivotal tutorial events in the game
# @param levels: Levels applicable for model, tut_features: tutorial features

CHECKPOINT_EVT = 2
extract_firsts = dict(sess_time_to_first_home_buy="first_house", sess_time_to_sell_food_tutorial="first_sale",
                      sess_time_to_first_livestock_buy="first_livestock", sess_time_to_first_farm_buy="first_farm",
                      sess_time_to_first_fertilizer_buy="first_fertilizer",
                      sess_time_to_extra_life_tutorial="first_death_of_last_farmbit")


class TutorialProgressionModel(FeatureModel):
    def __init__(self, tut_features: List[str]):
        super().__init__()
        self._tut_features = tut_features

    def _eval(self, sess: dict, verbose: bool = False):
        res = {}
        if sess['continue'] or sess['debug']:
            return None
        for x in self._tut_features:
            res[x] = sess[x]
        return res



__all__ = [
    "EventSchema",
    "ExtractorSchema",
    "DetectorSchema",
    "FeatureSchema",
    "AggregateSchema",
    "PerCountSchema"
]

from . import EventSchema
from . import ExtractorSchema
from . import DetectorSchema
from . import FeatureSchema
from . import AggregateSchema
from . import PerCountSchema


import re
import json


def main(path_to_lakeland_readme, outpath=None):
    """
    This function takes in a path to the lakeland readme and outputs a json file of all the enumerators to the outpath.
    This assumes event categories have not changed (as of 10/23/19) and the readme is up-to-date and has maintained its
    current format..
    :param path_to_lakeland_readme: str
    :param outpath: str
    :return: dict corresponding to the produced json
    """
    json_enums = {}
    table_enums = {}

    with open(path_to_lakeland_readme) as infile:
        enums = False
        table = ''
        write = False
        for line in infile:
            if line.strip() == "## Enumerators and Constants":
                enums = True
                continue
            elif line.strip() == "## Data Structures":
                enums = False
                json_enums[table] = table_enums
                table = ''
                write = False
                continue
            elif enums:
                print(table, write, line)
                if line.strip()[:4] == "####":
                    table = line.strip()[4:].strip().upper()
                    table_enums = {}
                    continue
                elif table and line.strip()[:2] == "<a":
                    json_enums[table] = table_enums
                    table = ''
                    write = False
                    continue
                elif table:
                    if line.strip() == "| --- | --- | --- |":
                        write = True
                    elif write:
                        if line.strip() and line.strip()[0] == "|":
                            matches = re.match(r'\|(.*)\|(.*)\|(.*)\|', line)
                            if len(matches.groups()) == 3:
                                idx = int(matches.groups()[0].strip())
                                val = matches.groups()[1].strip()
                                table_enums[val] = idx
    if outpath:
        with open(outpath,'w') as file:
            json.dump(json_enums, file, indent=2)
    return json_enums

if __name__ == '__main__':
    path = r"/Users/johnmccloskey/Development/lakeland/README.md"
    out = r"/Users/johnmccloskey/Development/opengamedata/Lakeland Enumerators/Lakeland Enumerators.json"
    main(path, out)





# import libraries
import json
from time import time
from typing import Any, Final, List, Optional
from datetime  import timedelta, datetime
# import local files
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event

class IdleState(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """

    IDLE_TIME_THRESHOLD : Final[timedelta] = timedelta(seconds=15)

    def __init__(self, params:GeneratorParameters, threshold:int):
        super().__init__(params=params)
        self._time : timedelta = timedelta(0)
        self._count : int = 0
        self._last_timestamp : Optional[datetime] = None
        self._threshold : timedelta = timedelta(seconds=threshold)

    @staticmethod
    def defaultThreshold():
        return IdleState.IDLE_TIME_THRESHOLD

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return [f"CUSTOM.{i}" for i in range(3, 21)] + ["CUSTOM.1"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "CUSTOM.1" and not self._last_timestamp:
            self._last_timestamp = event.Timestamp
            return
        if self._last_timestamp is not None:
            time_since_last = event.Timestamp - self._last_timestamp
            if time_since_last > IdleState.IDLE_TIME_THRESHOLD:
                self._time += time_since_last
                self._count += 1
        else:
            raise ValueError("In IdleState, last timestamp is None!")
        self._last_timestamp = event.Timestamp
        return

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._time, self._count]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["Count"] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    
    def BaseFeatureSuffix(self) -> str:
        return "Time"



# import libraries
from os import truncate
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models import Event
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.PerLevelFeature import PerLevelFeature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class ContinuesOnFail(SessionFeature):
    def __init__(self, params:GeneratorParameters):
        super().__init__(params=params)
        self._quit_type : Optional[bool] = None
        self._total_fails : Optional[int] = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    def _validateEventCountIndex(self, event:Event):
        
        return False
        #return int(event.GameState['att']) == self.CountIndex



    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return [""]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["QuitType", "TotalFails"]

    def _updateFromEvent(self, event:Event) -> None:
        #self._story_alignment = event.EventData["story_alignment"]

        pass


    def _updateFromFeatureData(self, feature:FeatureData):
        if(feature._name =="QuitType"):
            self._quit_type = feature._vals
        elif(feature._name == "TotalFails"):
            self._total_fails = feature._vals


            
        return

    def _getFeatureValues(self) -> List[Any]:
        
        if(self._quit_type[2] == False):
            #didn't quit on fail
            return [self._total_fails[0]]
        else:
            #quit on fail, so subtract 1 for 
            return [self._total_fails[0]-1]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return []
    
    @staticmethod
    def AvailableModes() -> List[ExtractionMode]:
        return [ExtractionMode.SESSION, ExtractionMode.PLAYER,ExtractionMode.POPULATION] # >>> delete any modes you don't want run for your Feature. <<<
    



# import standard libraries
import logging
# import locals
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.utils.Logger import Logger

class SessionFeature(Feature):

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, params:GeneratorParameters):
        if params._count_index is not None and params._count_index != 0:
            Logger.Log(f"Session feature {params._name} got non-zero count index of {params._count_index}!", logging.WARN)
        params._count_index = 0
        super().__init__(params=params)

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



## import standard libraries
import abc
from typing import Any, Dict

# import local files
from ogd.core.schemas.configs.GameSourceSchema import GameSourceSchema

class Interface(abc.ABC):

    # *** ABSTRACTS ***

    @abc.abstractmethod
    def _open(self) -> bool:
        pass

    @abc.abstractmethod
    def _close(self) -> bool:
        pass

    # *** BUILT-INS & PROPERTIES ***

    def __init__(self, config:GameSourceSchema):
        self._config  : GameSourceSchema = config
        self._is_open : bool = False

    def __del__(self):
        self.Close()

    # *** PUBLIC STATICS ***

    # *** PUBLIC METHODS ***

    def Open(self, force_reopen:bool = False) -> bool:
        if (not self._is_open) or force_reopen:
            self._is_open = self._open()
        return self._is_open
    
    def IsOpen(self) -> bool:
        return True if self._is_open else False

    def Close(self) -> bool:
        if self.IsOpen():
            return self._close()
        else:
            return True

    # *** PROPERTIES ***

    # *** PRIVATE STATICS ***

    # *** PRIVATE METHODS ***



# import libraries
import inspect
import logging
from datetime import timedelta
from typing import Any, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.Generator import GeneratorParameters
from ogd.games.AQUALAB.features.PerJobFeature import PerJobFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class JobCompletionTime(PerJobFeature):

    def __init__(self, params:GeneratorParameters, job_map:dict):
        super().__init__(params=params, job_map=job_map)
        self._session_id = None
        self._job_start_time = None
        self._prev_timestamp = None
        self._time = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["accept_job", "complete_job"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.SessionID != self._session_id:
            self._session_id = event.SessionID
            # if we jumped to a new session, we only want to count time up to last event, not the time between sessions.
            if self._job_start_time and self._prev_timestamp:
                self._time += (self._prev_timestamp - self._job_start_time).total_seconds()
                self._job_start_time = event.timestamp

        def _getFilename(full_path:str):
            return full_path.split('\\')[-1].split('.')[0]

        if event.event_name == "accept_job":
            self._job_start_time = event.timestamp
        elif event.event_name == "complete_job":
            if self._job_start_time:
                self._time += (event.timestamp - self._job_start_time).total_seconds()
                self._job_start_time = None
            else:
                _completed_job = event.GameState.get('job_name', event.EventData.get('job_name', "JOB NAME NOT FOUND"))
                callstack = [f"{_getFilename(inspect.stack()[i].filename)}.{inspect.stack()[i].function}" for i in range(min(11, len(inspect.stack())))]
                Logger.Log(f"In {callstack}:\n  {event.user_id} ({event.session_id}) completed job {_completed_job} with no active start time!", logging.DEBUG)

        self._prev_timestamp = event.timestamp

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [timedelta(seconds=self._time)]

    # *** Optionally override public functions. ***
    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



# import libraries
import logging
from typing import Any, Dict, List, Optional
# import locals
from ogd.core.utils.Logger import Logger
from ogd.core.generators.extractors.Extractor import GeneratorParameters
from ogd.core.generators.extractors.PerCountFeature import PerCountFeature
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData

class JobPriorAttempt(PerCountFeature):

    def __init__(self, params:GeneratorParameters, job_map:dict):
        self._job_map = job_map
        super().__init__(params=params)
        self._prior_list = set()
        self._completed = False
        self._session_id : Optional[int] = None

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["complete_job", "accept_job"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.app_version == 'Aqualab' or event.app_version == 'None':
            job_data = event.GameState.get("job_name", event.EventData.get("job_name", {})).get('string_value')
        else:
            job_data = event.GameState.get("job_name", event.EventData.get("job_name", {}))
        if event.event_name == "complete_job":
            if self._job_map[job_data] == self.CountIndex:
                self._completed = True
                return
        else:
            if self._job_map[job_data] != self.CountIndex:
                self._prior_list.add(self._job_map[job_data])
                return

    def _validateEventCountIndex(self, event:Event) -> bool:
        return not self._completed

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [sorted(self._prior_list)]

    # *** Optionally override public functions. ***

    def Subfeatures(self) -> List[str]:
        return []

    @staticmethod
    def MinVersion() -> Optional[str]:
        return "1"



## import standard libraries
import abc
import enum
import typing
import logging

@enum.unique
class ModelInputType(enum.Enum):
    FEATURE  = enum.auto()
    SEQUENCE = enum.auto()

    def __str__(self):
        return self.name

## @class Model
#  Abstract base class for models to be displayed in realtime dashboard.
#  Models only have one public function, called Eval.
#  The Eval function takes a list of row data, computes some statistic, and returns a list of results.
#  If the model works on features from session data, it should calculate one result for each row (each row being a session).
#  If the model works on a raw list of recent events, it should calculate a single result (each row being an event).
class Model(abc.ABC):
    def __init__(self, input_type: ModelInputType, levels: typing.List[int] = []):
        self._input_type : ModelInputType   = input_type
        self._levels     : typing.List[int] = levels

    def GetInputType(self):
        return self._input_type

    ## Abstract declaration of a function to perform calculation of a model results from a row.
    #
    #  @param rows : A list of rows, where each row contains all columns expected by the specific model implementation.
    #  @return     : A list of results for the data (how many results depends on the model)
    @abc.abstractmethod
    def Eval(self, rows: typing.List) -> typing.List:
        pass


# import libraries
from ogd.core.models import Event
from typing import Any, List, Optional
# import locals
from ogd.core.generators.extractors.Feature import Feature
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.models.Event import Event
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData


class PercentWavelengthGoodMoves(Feature):
    def __init__(self, params:GeneratorParameters):
        Feature.__init__(self, params=params)
        self._wavelength_count = 0
        self._good_count = 0

    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return ["CUSTOM.1", "CUSTOM.2"]
        # return ["SLIDER_MOVE_RELEASE", "ARROW_MOVE_RELEASE"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventData['slider'].upper() == 'WAVELENGTH':
            self._wavelength_count += 1
            if event.EventName == "CUSTOM.1":
                if event.EventData['end_closeness'] > event.EventData['begin_closeness']:
                    self._good_count += 1
            elif event.EventName == "CUSTOM.2":
                start_dist = event.EventData['correct_val'] - event.EventData['begin_val']
                end_dist = event.EventData['correct_val'] - event.EventData['end_val']
                if abs(end_dist) < abs(start_dist):
                    self._good_count += 1

    def _updateFromFeatureData(self, feature:FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._good_count / self._wavelength_count if self._wavelength_count != 0 else None]

    # *** Optionally override public functions. ***



# import libraries
import json
from typing import Any, Final, List, Optional
from datetime  import timedelta, datetime
# import local files
from ogd.core.generators.Generator import GeneratorParameters
from ogd.core.generators.extractors.SessionFeature import SessionFeature
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.models.FeatureData import FeatureData
from ogd.core.models.Event import Event

class ActiveStateTime(SessionFeature):
    """Template file to serve as a guide for creating custom Feature subclasses for games.

    :param Feature: Base class for a Custom Feature class.
    :type Feature: _type_
    """

    ACTIVE_TIME_THRESHOLD : Final[timedelta] = timedelta(seconds=15)
    CLICK_EVENTS_NAME     : Final[List[str]] = [f"CUSTOM.{i}" for i in range(3, 12)]

    def __init__(self, params:GeneratorParameters, threshold:int):
        super().__init__(params=params)
        self._time : timedelta = timedelta(0)
        self._clicking_time : timedelta = timedelta(0)
        self._last_hover_or_click_timestamp : Optional[datetime] = None
        self._last_click_timestamp : Optional[datetime] = None
        self._threshold = timedelta(seconds = threshold)


    # *** IMPLEMENT ABSTRACT FUNCTIONS ***
    @classmethod
    def _eventFilter(cls, mode:ExtractionMode) -> List[str]:
        return [f"CUSTOM.{i}" for i in range(3, 21)] + ["CUSTOM.1"]

    @classmethod
    def _featureFilter(cls, mode:ExtractionMode) -> List[str]:
        return []

    def _updateFromEvent(self, event:Event) -> None:
        if event.EventName == "CUSTOM.1" and not self._last_hover_or_click_timestamp:
            self._last_hover_or_click_timestamp = event.Timestamp
            self._last_click_timestamp = event.Timestamp
            return
        # elif event.EventName == "CUSTOM.1" or not self._last_hover_or_click_timestamp:
        #     raise(ValueError("Multiple game start events or none gamestart events!"))

        if self._last_hover_or_click_timestamp is not None:
            _time_between_hover_or_click = event.Timestamp - self._last_hover_or_click_timestamp
            if _time_between_hover_or_click < self._threshold:
                self._time += _time_between_hover_or_click
            self._last_hover_or_click_timestamp = event.Timestamp

        if self._last_click_timestamp is not None:
            if event.EventName in ActiveStateTime.CLICK_EVENTS_NAME:
                _time_between_click = event.Timestamp - self._last_click_timestamp
                if _time_between_click < self._threshold:
                    self._clicking_time += _time_between_click
                self._last_click_timestamp = event.Timestamp

        return

    def _updateFromFeatureData(self, feature: FeatureData):
        return

    def _getFeatureValues(self) -> List[Any]:
        return [self._time, self._clicking_time]

    # *** Optionally override public functions. ***
    def Subfeatures(self) -> List[str]:
        return ["Clicking"] # >>> fill in names of Subfeatures for which this Feature should extract values. <<<
    



# import standard libraries
import logging
from typing import Any, Dict, List, Set
# import local files
from ogd.core.models.enums.ExtractionMode import ExtractionMode
from ogd.core.schemas.Schema import Schema
from ogd.core.utils.Logger import Logger

class ColumnSchema(Schema):
    def __init__(self, all_elements:Dict[str, Any]):
        self._readable    : str
        self._value_type  : str
        self._description : str

        if not isinstance(all_elements, dict):
            self._name        = "No Name"
            self._readable    = "No Readable Name"
            self._description = "No Description"
            self._value_type  = "No Type"
            self._elements = {}
            Logger.Log(f"For {self._name} Extractor config, all_elements was not a dict, defaulting to empty dict", logging.WARN)
        if "readable" in all_elements.keys():
            self._readable = ColumnSchema._parseReadable(all_elements['readable'])
        else:
            self._readable = self._name
            Logger.Log(f"{self._name} config does not have a 'readable' element; defaulting to readable=name", logging.WARN)

        if "description" in all_elements.keys():
            self._description = ColumnSchema._parseDescription(all_elements['description'])
        else:
            self._description = ""
            Logger.Log(f"{self._name} config does not have an 'description' element; defaulting to description=''", logging.WARN)

        if "type" in all_elements.keys():
            self._value_type = ColumnSchema._parseValueType(all_elements['type'])
        else:
            self._value_type = self._name

        _name : str
        if "name" in all_elements.keys():
            _name = ColumnSchema._parseName(all_elements['name'])
        else:
            _name = "NOT FOUND"
            Logger.Log(f"Column config does not have a 'name' element; defaulting to name=NOT FOUND", logging.WARN)
        _leftovers = { key : val for key,val in all_elements.items() if key not in {"name", "readable", "description", "type"} }
        super().__init__(name=_name, other_elements=_leftovers)

    def __str__(self):
        return self.Name

    def __repr__(self):
        return self.Name

    @property
    def ReadableName(self) -> str:
        return self._name

    @property
    def Description(self) -> str:
        return self._description

    @property
    def ValueType(self) -> str:
        return self._value_type

    @property
    def AsMarkdown(self) -> str:
        ret_val = f"**{self.Name}** : *{self.ValueType}* - {self.ReadableName}, {self.Description}  "

        if len(self.NonStandardElements) > 0:
            other_elems = [f"{key}: {val}" for key,val in self.NonStandardElements]
            ret_val += f"\n    Other Elements: {', '.join(other_elems)}"

        return ret_val
    
    @staticmethod
    def _parseName(name):
        ret_val : str
        if isinstance(name, str):
            ret_val = name
        else:
            ret_val = str(name)
            Logger.Log(f"Column name was not a string, defaulting to str(name) == {ret_val}", logging.WARN)
        return ret_val
    
    @staticmethod
    def _parseReadable(name):
        ret_val : str
        if isinstance(name, str):
            ret_val = name
        else:
            ret_val = str(name)
            Logger.Log(f"Column readable name was not a string, defaulting to str(readable) == {ret_val}", logging.WARN)
        return ret_val
    
    @staticmethod
    def _parseDescription(description):
        ret_val : str
        if isinstance(description, str):
            ret_val = description
        else:
            ret_val = str(description)
            Logger.Log(f"Column description was not a string, defaulting to str(description) == {ret_val}", logging.WARN)
        return ret_val
    
    @staticmethod
    def _parseValueType(extractor_type):
        ret_val : str
        if isinstance(extractor_type, str):
            ret_val = extractor_type
        else:
            ret_val = str(extractor_type)
            Logger.Log(f"Column type was not a string, defaulting to str(type) == {ret_val}", logging.WARN)
        return ret_val
